@@ -21,9 +21,16 @@
 #include "gtkscalerprivate.h"
 
 #include "gdk/gdktextureprivate.h"
+#include "gdk/loaders/gdkpngprivate.h"
 
 /* {{{ Pixbuf helpers */
 
+static inline gboolean
+pixbuf_is_only_fg (GdkPixbuf *pixbuf)
+{
+  return gdk_pixbuf_get_option (pixbuf, "tEXt::only-foreground") != NULL;
+}
+
 static GdkPixbuf *
 load_from_stream (GdkPixbufLoader  *loader,
                   GInputStream     *stream,
@@ -199,6 +206,7 @@ _gdk_pixbuf_new_from_resource_at_scale (const char   *resource_path,
 
 static GdkPixbuf *
 load_symbolic_svg (const char     *escaped_file_data,
+                   gsize           len,
                    int             width,
                    int             height,
                    const char     *icon_width_str,
@@ -214,37 +222,38 @@ load_symbolic_svg (const char     *escaped_file_data,
   char *data;
 
   data = g_strconcat ("<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n"
-                      "<svg version=\"1.1\"\n"
-                      "     xmlns=\"http://www.w3.org/2000/svg\"\n"
-                      "     xmlns:xi=\"http://www.w3.org/2001/XInclude\"\n"
-                      "     width=\"", icon_width_str, "\"\n"
-                      "     height=\"", icon_height_str, "\">\n"
-                      "  <style type=\"text/css\">\n"
-                      "    rect,circle,path {\n"
-                      "      fill: ", fg_string," !important;\n"
-                      "    }\n"
-                      "    .warning {\n"
-                      "      fill: ", warning_color_string, " !important;\n"
-                      "    }\n"
-                      "    .error {\n"
-                      "      fill: ", error_color_string ," !important;\n"
-                      "    }\n"
-                      "    .success {\n"
-                      "      fill: ", success_color_string, " !important;\n"
-                      "    }\n"
-                      "  </style>\n"
-                      "  <xi:include href=\"data:text/xml;base64,", escaped_file_data, "\"/>\n"
-                      "</svg>",
+                      "<svg version=\"1.1\" "
+                           "xmlns=\"http://www.w3.org/2000/svg\" "
+                           "xmlns:xi=\"http://www.w3.org/2001/XInclude\" "
+                           "width=\"", icon_width_str, "\" "
+                           "height=\"", icon_height_str, "\">"
+                        "<style type=\"text/css\">"
+                          "rect,circle,path {"
+                            "fill: ", fg_string," !important;"
+                          "}\n"
+                          ".warning {"
+                             "fill: ", warning_color_string, " !important;"
+                          "}\n"
+                          ".error {"
+                            "fill: ", error_color_string ," !important;"
+                          "}\n"
+                          ".success {"
+                            "fill: ", success_color_string, " !important;"
+                          "}"
+                        "</style>"
+                        "<xi:include href=\"data:text/xml;base64,",
                       NULL);
 
   stream = g_memory_input_stream_new_from_data (data, -1, g_free);
+  g_memory_input_stream_add_data (G_MEMORY_INPUT_STREAM (stream), escaped_file_data, len, NULL);
+  g_memory_input_stream_add_data (G_MEMORY_INPUT_STREAM (stream), "\"/></svg>", strlen ("\"/></svg>"), NULL);
   pixbuf = gdk_pixbuf_new_from_stream_at_scale (stream, width, height, TRUE, NULL, error);
   g_object_unref (stream);
 
   return pixbuf;
 }
 
-static void
+static gboolean
 extract_plane (GdkPixbuf *src,
                GdkPixbuf *dst,
                int        from_plane,
@@ -255,6 +264,7 @@ extract_plane (GdkPixbuf *src,
   gsize src_stride, dst_stride;
   guchar *src_row, *dst_row;
   int x, y;
+  gboolean all_clear = TRUE;
 
   width = gdk_pixbuf_get_width (src);
   height = gdk_pixbuf_get_height (src);
@@ -274,11 +284,109 @@ extract_plane (GdkPixbuf *src,
       dst_row = dst_data + dst_stride * y;
       for (x = 0; x < width; x++)
         {
+          if (src_row[from_plane] != 0)
+            all_clear = FALSE;
+
           dst_row[to_plane] = src_row[from_plane];
           src_row += 4;
           dst_row += 4;
         }
     }
+
+  return all_clear;
+}
+
+static void
+keep_alpha (GdkPixbuf *src)
+{
+  guchar *data;
+  int width, height;
+  gsize stride;
+
+  data = gdk_pixbuf_get_pixels (src);
+  width = gdk_pixbuf_get_width (src);
+  height = gdk_pixbuf_get_height (src);
+  stride = gdk_pixbuf_get_rowstride (src);
+
+  for (int y = 0; y < height; y++)
+    {
+      guchar *row = data + stride * y;
+      for (int x = 0; x < width; x++)
+        {
+          row[0] = row[1] = row[2] = 0;
+          row += 4;
+        }
+    }
+}
+
+static void
+svg_find_size_strings (const char  *data,
+                       gsize        len,
+                       char       **width,
+                       char       **height)
+{
+  gsize i, j, k, l;
+
+  *width = NULL;
+  *height = NULL;
+
+  for (i = 0; i < len - 4; i++)
+    {
+      if (strncmp (data + i, "<svg", 4) == 0)
+        {
+          for (j = i + strlen ("<svg"); j < len - 9; j++)
+            {
+              if (strncmp (data + j, "height=\"", strlen ("height=\"")) == 0)
+                {
+                  k = l = j + strlen ("height=\"");
+                  while (l < len && data[l] != '\"')
+                    l++;
+
+                  *height = g_strndup (data + k, l - k);
+
+                  if (*width && *height)
+                    return;
+
+                  j = l;
+                }
+              else if (strncmp (data + j, "width=\"", strlen ("width=\"")) == 0)
+                {
+                  k = l = j + strlen ("width=\"");
+                  while (l < len && data[l] != '\"')
+                    l++;
+
+                  *width = g_strndup (data + k, l - k);
+
+                  if (*width && *height)
+                    return;
+
+                  j = l;
+                }
+              else if (data[j] == '>')
+                {
+                  break;
+                }
+            }
+
+          break;
+        }
+    }
+
+  *width = g_strdup ("16px");
+  *height = g_strdup ("16px");
+}
+
+static gboolean
+svg_has_symbolic_classes (const char *data,
+                          gsize       len)
+{
+#ifdef HAVE_MEMMEM
+  return memmem (data, len, "class=\"error\"", strlen ("class=\"error\"")) != NULL ||
+         memmem (data, len, "class=\"warning\"", strlen ("class=\"warning\"")) != NULL ||
+         memmem (data, len, "class=\"success\"", strlen ("class=\"success\"")) != NULL;
+#else
+  return TRUE;
+#endif
 }
 
 GdkPixbuf *
@@ -293,38 +401,46 @@ gtk_make_symbolic_pixbuf_from_data (const char  *file_data,
 {
   const char *r_string = "rgb(255,0,0)";
   const char *g_string = "rgb(0,255,0)";
-  char *icon_width_str;
-  char *icon_height_str;
-  GdkPixbuf *loaded;
+  char *icon_width_str = NULL;
+  char *icon_height_str = NULL;
+  char *escaped_file_data = NULL;
+  gsize len;
   GdkPixbuf *pixbuf = NULL;
-  int plane;
-  int icon_width, icon_height;
-  char *escaped_file_data;
+  gboolean has_symbolic_classes;
+  gboolean only_fg = TRUE;
+
+  has_symbolic_classes = svg_has_symbolic_classes (file_data, file_len);
 
   /* Fetch size from the original icon */
-  GInputStream *stream = g_memory_input_stream_new_from_data (file_data, file_len, NULL);
-  GdkPixbuf *reference = gdk_pixbuf_new_from_stream (stream, NULL, error);
-
-  g_object_unref (stream);
-
-  if (!reference)
-    return NULL;
-
-  icon_width = gdk_pixbuf_get_width (reference);
-  icon_height = gdk_pixbuf_get_height (reference);
-  g_object_unref (reference);
-
-  escaped_file_data = g_base64_encode ((guchar *) file_data, file_len);
-  icon_width_str = g_strdup_printf ("%d", icon_width);
-  icon_height_str = g_strdup_printf ("%d", icon_height);
+  if (has_symbolic_classes || width == 0 || height == 0)
+    svg_find_size_strings (file_data, file_len, &icon_width_str, &icon_height_str);
 
   if (width == 0)
-    width = icon_width * scale;
+    width = (int) (g_ascii_strtoull (icon_width_str, NULL, 0) * scale);
   if (height == 0)
-    height = icon_height * scale;
+    height = (int) (g_ascii_strtoull (icon_height_str, NULL, 0) * scale);
 
-  for (plane = 0; plane < 3; plane++)
+  if (!has_symbolic_classes)
     {
+      GInputStream *stream;
+
+      stream = g_memory_input_stream_new_from_data (file_data, file_len, NULL);
+      pixbuf = gdk_pixbuf_new_from_stream_at_scale (stream, width, height, TRUE, NULL, error);
+      g_object_unref (stream);
+
+      if (pixbuf)
+        keep_alpha (pixbuf);
+
+      goto out;
+    }
+
+  escaped_file_data = g_base64_encode ((guchar *) file_data, file_len);
+  len = strlen (escaped_file_data);
+
+  for (int plane = 0; plane < 3; plane++)
+    {
+      GdkPixbuf *loaded;
+
       /* Here we render the svg with all colors solid, this should
        * always make the alpha channel the same and it should match
        * the final alpha channel for all possible renderings. We
@@ -337,7 +453,7 @@ gtk_make_symbolic_pixbuf_from_data (const char  *file_data,
        * channels, with the color of the fg being implicitly
        * the "rest", as all color fractions should add up to 1.
        */
-      loaded = load_symbolic_svg (escaped_file_data, width, height,
+      loaded = load_symbolic_svg (escaped_file_data, len, width, height,
                                   icon_width_str,
                                   icon_height_str,
                                   g_string,
@@ -358,25 +474,26 @@ gtk_make_symbolic_pixbuf_from_data (const char  *file_data,
           g_free (filename);
         }
 
-      if (pixbuf == NULL)
+      if (plane == 0)
         {
           pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8,
                                    gdk_pixbuf_get_width (loaded),
                                    gdk_pixbuf_get_height (loaded));
-          gdk_pixbuf_fill (pixbuf, 0);
+          memset (gdk_pixbuf_get_pixels (pixbuf), 0, gdk_pixbuf_get_byte_length (pixbuf));
+
+          extract_plane (loaded, pixbuf, 3, 3);
         }
 
-      if (plane == 0)
-        extract_plane (loaded, pixbuf, 3, 3);
-
-      extract_plane (loaded, pixbuf, 0, plane);
+      only_fg &= extract_plane (loaded, pixbuf, 0, plane);
 
       g_object_unref (loaded);
     }
 
-  g_free (escaped_file_data);
-
 out:
+  if (only_fg && pixbuf)
+    gdk_pixbuf_set_option (pixbuf, "tEXt::only-foreground", "true");
+
+  g_free (escaped_file_data);
   g_free (icon_width_str);
   g_free (icon_height_str);
 
@@ -409,17 +526,17 @@ make_symbolic_pixbuf_from_resource (const char  *path,
 }
 
 static GdkPixbuf *
-make_symbolic_pixbuf_from_path (const char  *path,
-                                int          width,
-                                int          height,
-                                double       scale,
-                                GError     **error)
+make_symbolic_pixbuf_from_filename (const char  *filename,
+                                    int          width,
+                                    int          height,
+                                    double       scale,
+                                    GError     **error)
 {
   char *data;
   gsize size;
   GdkPixbuf *pixbuf;
 
-  if (!g_file_get_contents (path, &data, &size, error))
+  if (!g_file_get_contents (filename, &data, &size, error))
     return NULL;
 
   pixbuf = gtk_make_symbolic_pixbuf_from_data (data, size, width, height, scale, NULL, error);
@@ -453,11 +570,85 @@ make_symbolic_pixbuf_from_file (GFile       *file,
 /* }}} */
 /* {{{ Texture API */
 
+static GdkTexture *
+texture_new_from_bytes (GBytes    *bytes,
+                        gboolean  *only_fg,
+                        GError   **error)
+{
+  GHashTable *options;
+  GdkTexture *texture;
+
+  if (!gdk_is_png (bytes))
+    return gdk_texture_new_from_bytes (bytes, error);
+
+  options = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
+  texture = gdk_load_png (bytes, options, error);
+  *only_fg = g_hash_table_contains (options, "foreground-only");
+  g_hash_table_unref (options);
+
+  return texture;
+}
+
+GdkTexture *
+gdk_texture_new_from_filename_with_fg (const char    *filename,
+                                       gboolean      *only_fg,
+                                       GError       **error)
+{
+  GFile *file;
+  GBytes *bytes;
+  GdkTexture *texture = NULL;
+
+  file = g_file_new_for_path (filename);
+  bytes = g_file_load_bytes (file, NULL, NULL, error);
+  if (bytes)
+    texture = texture_new_from_bytes (bytes, only_fg, error);
+  g_bytes_unref (bytes);
+  g_object_unref (file);
+
+  return texture;
+}
+
+GdkTexture *
+gdk_texture_new_from_resource_with_fg (const char *path,
+                                       gboolean   *only_fg)
+{
+  GBytes *bytes;
+  GdkTexture *texture = NULL;
+
+  bytes = g_resources_lookup_data (path, 0, NULL);
+  if (bytes)
+    texture = texture_new_from_bytes (bytes, only_fg, NULL);
+  g_bytes_unref (bytes);
+
+  return texture;
+}
+
+GdkTexture *
+gdk_texture_new_from_stream_with_fg (GInputStream  *stream,
+                                     gboolean      *only_fg,
+                                     GCancellable  *cancellable,
+                                     GError       **error)
+{
+  GdkPixbuf *pixbuf;
+  GdkTexture *texture = NULL;
+
+  pixbuf = _gdk_pixbuf_new_from_stream_scaled (stream, 0, cancellable, error);
+  if (pixbuf)
+    {
+      *only_fg = pixbuf_is_only_fg (pixbuf);
+      texture = gdk_texture_new_for_pixbuf (pixbuf);
+      g_object_unref (pixbuf);
+    }
+
+  return texture;
+}
+
 GdkTexture *
 gdk_texture_new_from_stream_at_scale (GInputStream  *stream,
                                       int            width,
                                       int            height,
                                       gboolean       aspect,
+                                      gboolean      *only_fg,
                                       GCancellable  *cancellable,
                                       GError       **error)
 {
@@ -467,24 +658,7 @@ gdk_texture_new_from_stream_at_scale (GInputStream  *stream,
   pixbuf = _gdk_pixbuf_new_from_stream_at_scale (stream, width, height, aspect, cancellable, error);
   if (pixbuf)
     {
-      texture = gdk_texture_new_for_pixbuf (pixbuf);
-      g_object_unref (pixbuf);
-    }
-
-  return texture;
-}
-
-GdkTexture *
-gdk_texture_new_from_stream (GInputStream  *stream,
-                             GCancellable  *cancellable,
-                             GError       **error)
-{
-  GdkPixbuf *pixbuf;
-  GdkTexture *texture = NULL;
-
-  pixbuf = _gdk_pixbuf_new_from_stream_scaled (stream, 0, cancellable, error);
-  if (pixbuf)
-    {
+      *only_fg = pixbuf_is_only_fg (pixbuf);
       texture = gdk_texture_new_for_pixbuf (pixbuf);
       g_object_unref (pixbuf);
     }
@@ -497,6 +671,7 @@ gdk_texture_new_from_resource_at_scale (const char    *path,
                                         int            width,
                                         int            height,
                                         gboolean       preserve_aspect,
+                                        gboolean      *only_fg,
                                         GError       **error)
 {
   GdkPixbuf *pixbuf;
@@ -505,6 +680,7 @@ gdk_texture_new_from_resource_at_scale (const char    *path,
   pixbuf = _gdk_pixbuf_new_from_resource_at_scale (path, width, height, preserve_aspect, error);
   if (pixbuf)
     {
+      *only_fg = pixbuf_is_only_fg (pixbuf);
       texture = gdk_texture_new_for_pixbuf (pixbuf);
       g_object_unref (pixbuf);
     }
@@ -516,18 +692,20 @@ gdk_texture_new_from_resource_at_scale (const char    *path,
 /* {{{ Symbolic texture API */
 
 GdkTexture *
-gdk_texture_new_from_path_symbolic (const char    *path,
-                                    int            width,
-                                    int            height,
-                                    double         scale,
-                                    GError       **error)
+gdk_texture_new_from_filename_symbolic (const char    *filename,
+                                        int            width,
+                                        int            height,
+                                        double         scale,
+                                        gboolean      *only_fg,
+                                        GError       **error)
 {
   GdkPixbuf *pixbuf;
   GdkTexture *texture = NULL;
 
-  pixbuf = make_symbolic_pixbuf_from_path (path, width, height, scale, error);
+  pixbuf = make_symbolic_pixbuf_from_filename (filename, width, height, scale, error);
   if (pixbuf)
     {
+      *only_fg = pixbuf_is_only_fg (pixbuf);
       texture = gdk_texture_new_for_pixbuf (pixbuf);
       g_object_unref (pixbuf);
     }
@@ -546,6 +724,7 @@ gdk_texture_new_from_resource_symbolic (const char  *path,
                                         int          width,
                                         int          height,
                                         double       scale,
+                                        gboolean    *only_fg,
                                         GError     **error)
 {
   GdkPixbuf *pixbuf;
@@ -554,6 +733,7 @@ gdk_texture_new_from_resource_symbolic (const char  *path,
   pixbuf = make_symbolic_pixbuf_from_resource (path, width, height, scale, error);
   if (pixbuf)
     {
+      *only_fg = pixbuf_is_only_fg (pixbuf);
       texture = gdk_texture_new_for_pixbuf (pixbuf);
       g_object_unref (pixbuf);
     }
@@ -588,14 +768,19 @@ gdk_texture_new_from_file_symbolic (GFile       *file,
                                     int          width,
                                     int          height,
                                     double       scale,
+                                    gboolean    *only_fg,
                                     GError     **error)
 {
   GdkPixbuf *pixbuf;
-  GdkTexture *texture;
+  GdkTexture *texture = NULL;
 
   pixbuf = make_symbolic_pixbuf_from_file (file, width, height, scale, error);
-  texture = gdk_texture_new_for_pixbuf (pixbuf);
-  g_object_unref (pixbuf);
+  if (pixbuf)
+    {
+      *only_fg = pixbuf_is_only_fg (pixbuf);
+      texture = gdk_texture_new_for_pixbuf (pixbuf);
+      g_object_unref (pixbuf);
+    }
 
   return texture;
 }
@@ -604,7 +789,7 @@ gdk_texture_new_from_file_symbolic (GFile       *file,
 /* {{{ Scaled paintable API */
 
 typedef struct {
-  int scale_factor;
+  double scale;
 } LoaderData;
 
 static void
@@ -620,24 +805,24 @@ on_loader_size_prepared (GdkPixbufLoader *loader,
   format = gdk_pixbuf_loader_get_format (loader);
   if (!gdk_pixbuf_format_is_scalable (format))
     {
-      loader_data->scale_factor = 1;
+      loader_data->scale = 1.0;
       return;
     }
 
   gdk_pixbuf_loader_set_size (loader,
-                              width * loader_data->scale_factor,
-                              height * loader_data->scale_factor);
+                              width * loader_data->scale,
+                              height * loader_data->scale);
 }
 
 static GdkPaintable *
 gdk_paintable_new_from_bytes_scaled (GBytes *bytes,
-                                     int     scale_factor)
+                                     double  scale)
 {
   LoaderData loader_data;
   GdkTexture *texture;
   GdkPaintable *paintable;
 
-  loader_data.scale_factor = scale_factor;
+  loader_data.scale = scale;
 
   if (gdk_texture_can_load (bytes))
     {
@@ -667,8 +852,8 @@ gdk_paintable_new_from_bytes_scaled (GBytes *bytes,
       texture = gdk_texture_new_for_pixbuf (gdk_pixbuf_loader_get_pixbuf (loader));
       g_object_unref (loader);
 
-      if (loader_data.scale_factor != 1)
-        paintable = gtk_scaler_new (GDK_PAINTABLE (texture), loader_data.scale_factor);
+      if (loader_data.scale != 1.0)
+        paintable = gtk_scaler_new (GDK_PAINTABLE (texture), loader_data.scale);
       else
         paintable = g_object_ref (GDK_PAINTABLE (texture));
 
@@ -679,20 +864,20 @@ gdk_paintable_new_from_bytes_scaled (GBytes *bytes,
 }
 
 GdkPaintable *
-gdk_paintable_new_from_path_scaled (const char *path,
-                                    int         scale_factor)
+gdk_paintable_new_from_filename_scaled (const char *filename,
+                                        double      scale)
 {
   char *contents;
   gsize length;
   GBytes *bytes;
   GdkPaintable *paintable;
 
-  if (!g_file_get_contents (path, &contents, &length, NULL))
+  if (!g_file_get_contents (filename, &contents, &length, NULL))
     return NULL;
 
   bytes = g_bytes_new_take (contents, length);
 
-  paintable = gdk_paintable_new_from_bytes_scaled (bytes, scale_factor);
+  paintable = gdk_paintable_new_from_bytes_scaled (bytes, scale);
 
   g_bytes_unref (bytes);
 
@@ -701,7 +886,7 @@ gdk_paintable_new_from_path_scaled (const char *path,
 
 GdkPaintable *
 gdk_paintable_new_from_resource_scaled (const char *path,
-                                        int         scale_factor)
+                                        double      scale)
 {
   GBytes *bytes;
   GdkPaintable *paintable;
@@ -710,15 +895,16 @@ gdk_paintable_new_from_resource_scaled (const char *path,
   if (!bytes)
     return NULL;
 
-  paintable = gdk_paintable_new_from_bytes_scaled (bytes, scale_factor);
+  paintable = gdk_paintable_new_from_bytes_scaled (bytes, scale);
+
   g_bytes_unref (bytes);
 
   return paintable;
 }
 
 GdkPaintable *
-gdk_paintable_new_from_file_scaled (GFile *file,
-                                    int    scale_factor)
+gdk_paintable_new_from_file_scaled (GFile  *file,
+                                    double  scale)
 {
   GBytes *bytes;
   GdkPaintable *paintable;
@@ -727,7 +913,7 @@ gdk_paintable_new_from_file_scaled (GFile *file,
   if (!bytes)
     return NULL;
 
-  paintable = gdk_paintable_new_from_bytes_scaled (bytes, scale_factor);
+  paintable = gdk_paintable_new_from_bytes_scaled (bytes, scale);
 
   g_bytes_unref (bytes);
 
@@ -736,4 +922,4 @@ gdk_paintable_new_from_file_scaled (GFile *file,
 
 /* }}} */
 
-/* vim:set foldmethod=marker expandtab: */
+/* vim:set foldmethod=marker: */
